#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "help_mp.h"
#include "mp_msg.h"

#include "aviheader.h"
#include "ms_hdr.h"

#include "muxer.h"

#include "oggpcm.h"
#include "ogguvs.h"
struct ogg_stream_info {
    int initialized;
    oggpcm_stream_t pcm_stream;
    ogguvs_stream_t uvs_stream;
};

static muxer_stream_t* oggfile_new_stream(muxer_t *muxer,int type){
    struct ogg_stream_info *si;
    muxer_stream_t* s;
    if(!muxer) return NULL;
    if(muxer->avih.dwStreams>=MUXER_MAX_STREAMS) {
	    mp_msg(MSGT_MUXER, MSGL_ERR, "Too many streams! increase MUXER_MAX_STREAMS !\n");
	    return NULL;
    }
    
    /* Initialze muxer stream info */
    s=malloc(sizeof(muxer_stream_t));
    if(!s) return NULL; // no mem!?
    memset(s,0,sizeof(muxer_stream_t));
    muxer->streams[muxer->avih.dwStreams]=s;
    s->type=type;
    s->id=muxer->avih.dwStreams;
    s->timer=0.0;
    s->size=0;
    s->muxer=muxer;
    s->priv=si=malloc(sizeof(struct ogg_stream_info));
    
    /* Initialize private data */
    memset(si,0,sizeof(struct ogg_stream_info));
    si->initialized = 0;

    switch(type){
    case MUXER_TYPE_VIDEO:
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_new_stream MUXER_TYPE_VIDEO\n");
      s->ckid=mmioFOURCC(('0'+s->id/10),('0'+(s->id%10)),'d','c');
      s->h.fccType=streamtypeVIDEO;
      if(!muxer->def_v) muxer->def_v=s;
      break;
    case MUXER_TYPE_AUDIO:
      mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_new_stream MUXER_TYPE_AUDIO\n");
      s->ckid=mmioFOURCC(('0'+s->id/10),('0'+(s->id%10)),'w','b');
      s->h.fccType=streamtypeAUDIO;
      break;
    default:
      mp_msg(MSGT_MUXER, MSGL_WARN, "Warning! unknown stream type: %d\n",type);
      return NULL;
    }
    muxer->avih.dwStreams++;
    return s;
}

static void oggfile_write_chunk(muxer_stream_t *s,size_t len,unsigned int flags, double dts, double pts){
    muxer_t *muxer=s->muxer;
    struct ogg_stream_info *si = s->priv;
    static int dbgframe=0;
       (void)dts; (void)pts;
    
    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_write_chunk %d - %.5lfs\n",
    dbgframe++, s->timer);
    
    // write out the chunk:
    if (s->type==MUXER_TYPE_AUDIO && si->initialized) {
        if(!oggpcm_stream_packetin (&si->pcm_stream, s->buffer, len)) {
            // Put the packet into the stream, now write out pages.
            oggpcm_stream_pageout(&si->pcm_stream, muxer->file);
        }        
    }
    
    if (s->type==MUXER_TYPE_VIDEO && si->initialized) {
         if(len < si->uvs_stream.field_size) {
            mp_msg(MSGT_MUXER, MSGL_V, 
                "[MUXER_OGG] Dropping packet: len(%d) < field size(%d)\n",
                len, si->uvs_stream.field_size);
        } else if(len > si->uvs_stream.field_size) {
            mp_msg(MSGT_MUXER, MSGL_V, 
                "[MUXER_OGG] Warning: len(%d) > field size(%d),"
                " using field size\n", len, si->uvs_stream.field_size
                );
            len = si->uvs_stream.field_size;
        }
        if(si->uvs_stream.fld_rate_num) {
            /* Fixed frame rate */
            if(!ogguvs_stream_packetin (&si->uvs_stream, OGGUVS_FLD_TOP, s->buffer,len, 0.0)) {
                // Put the packet into the stream, now write out pages.
                ogguvs_stream_pageout(&si->uvs_stream, muxer->file);
            }
        } else {
            /* Variable frame rate */
            if(!ogguvs_stream_packetin (&si->uvs_stream, OGGUVS_FLD_TOP,
            s->buffer,len, s->timer)) {
                // Put the packet into the stream, now write out pages.
                ogguvs_stream_pageout(&si->uvs_stream, muxer->file);
            }
        }                    
    }
    
/*    // alter counters:
    if(s->h.dwSampleSize){
        // CBR
        s->h.dwLength+=len/s->h.dwSampleSize;
        if(len%s->h.dwSampleSize)
            mp_msg(MSGT_MUXER,MSGL_WARN,MSGTR_WarningLenIsntDivisible);
    } else {
        // VBR
        s->h.dwLength++;
    }
    s->timer=(double)s->h.dwLength*s->h.dwScale/s->h.dwRate;
    s->size+=len;
*/
}

static void oggfile_write_header(muxer_t *muxer){
    unsigned int i;
    muxer_stream_t *s;
    struct ogg_stream_info *si;

    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_write_header\n");
    for(i=0;i<muxer->avih.dwStreams;i++){
        s = muxer->streams[i];
        si = s->priv;
        switch(s->type){
        case MUXER_TYPE_VIDEO:
        {
            int result;
            ogguvs_format_t fmt = OGGUVS_FMT_OTHER;
            ogguvs_colorsp_t csp = OGGUVS_CSP_UNSPEC_YCBCR;
            long img_size =0;
            
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG] Stream %d - VIDEO\n",i);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Size:        %d\n",
              s->bih->biSize);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Width:       %d\n",
              s->bih->biWidth);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Height:      %d\n",
              s->bih->biHeight);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Planes:      %d\n",
              s->bih->biPlanes);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   BitCount:    %d\n",
              s->bih->biBitCount);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   Compression: 0x%08X\n",
              s->bih->biCompression);
            mp_msg(MSGT_MUXER, MSGL_INFO, "[MUXER_OGG]   SizeImage:   %d\n",
              s->bih->biSizeImage);
            
            if(si->initialized) {
                mp_msg(MSGT_MUXER, MSGL_V,
                    "[MUXER_OGG] Writing stream footer.\n");
                if(ogguvs_footer_pageout(&si->uvs_stream, muxer->file)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                            "Failed to write OggUVS EOS packet.\n");
                    result = 0;
                }
            } else {
                mp_msg(MSGT_MUXER, MSGL_V,
                    "[MUXER_OGG] Writing stream header.\n");

                switch(s->bih->biCompression) {
                  case 0x32315659:
                      fmt = OGGUVS_FMT_YV12;
                      img_size = s->bih->biWidth * s->bih->biHeight * 3/2;
                      break;
                  case 0x42475220:
                      if(s->bih->biBitCount == 32) {
                          /* No way to tell if this is ARGB or just RGB32 */
                          fmt = OGGUVS_FMT_ARGBDIB;
                          img_size = s->bih->biWidth * s->bih->biHeight * 4;
                      }
                      break;
 
                }
            
                result = (fmt != OGGUVS_FMT_OTHER);
                if(result)
                    result = !si->initialized;
                else
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                        "Video format not supported in this container, "
                        "ignoring stream. (0x%08X)\n", s->bih->biCompression);

                { 
                    double fps=(float)s->h.dwRate/(float)s->h.dwScale;                    
                    fps = (fps < 1000.0)?fps:0;
                    
                    if(result && ogguvs_stream_init(&si->uvs_stream,
                                                     s->bih->biWidth,
                                                     s->bih->biHeight,
                                                     1.0,
                                                     fps,
                                                     90000,
                                                     img_size,
                                                     csp,
                                                     fmt,
                                                     s->ckid)) {
                        mp_msg(MSGT_MUXER, MSGL_WARN,
                            "OggUVS stream initialization failed, "
                            "ignoring stream.\n");
                        result = 0;
                    }
                }
                if(result && ogguvs_header_pageout(&si->uvs_stream, muxer->file)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                            "Failed to write OggUVS header to output file, "
                            "ignoring stream.\n");
                    result = 0;
                }

                if(result)
                    si->initialized = 1;
            }
        }
	    break;
        case MUXER_TYPE_AUDIO:
        {
            int result;
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] Stream %d - AUDIO\n",i);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   wFormatTag:      0x%08X\n",
              s->wf->wFormatTag);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nChannels:       %d\n",
              s->wf->nChannels);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nSamplesPerSec:  %d\n",
              s->wf->nSamplesPerSec);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nAvgBytesPerSec: %d\n",
              s->wf->nAvgBytesPerSec);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   nBlockAlign:     %d\n",
              s->wf->nBlockAlign);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   wBitsPerSample:  %d\n",
              s->wf->wBitsPerSample);
            mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG]   cbSize:          %d\n",
              s->wf->cbSize);
            
            if(si->initialized) {
                mp_msg(MSGT_MUXER, MSGL_V,
                    "[MUXER_OGG] Writing stream footer.\n");
                if(oggpcm_footer_pageout(&si->pcm_stream, muxer->file)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                            "Failed to write OggPCM EOS packet.\n");
                    result = 0;
                }

            } else {
                mp_msg(MSGT_MUXER, MSGL_V,
                    "[MUXER_OGG] Writing stream header.\n");

                result = (s->wf->wFormatTag == 1 && s->wf->wBitsPerSample == 16);
                if(result)
                    result = !si->initialized;
                else
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                        "Audio format not supported in this container, "
                        "ignoring stream.\n");
                        
                if(result && oggpcm_stream_init(&si->pcm_stream,
                                                 s->wf->nChannels,
                                                 s->wf->wBitsPerSample,
                                                 s->wf->nSamplesPerSec,
                                                 OGGPCM_FMT_S16_LE,
                                                 s->ckid)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                        "OggPCM stream initialization failed, "
                        "ignoring stream.\n");
                    result = 0;
                }

                if(result && oggpcm_header_pageout(&si->pcm_stream, muxer->file)) {
                    mp_msg(MSGT_MUXER, MSGL_WARN,
                            "Failed to write OggPCM header to output file, "
                            "ignoring stream.\n");
                    result = 0;
                }

                if(result)
                    si->initialized = 1;
            }
        }	  
	    break;
        }
    }
    return;
}

static void oggfile_write_index(muxer_t *muxer){
    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] oggfile_write_index\n");
    muxer->cont_new_stream = &oggfile_new_stream;
       return;
}

int muxer_init_muxer_ogg(muxer_t *muxer){
    mp_msg(MSGT_MUXER, MSGL_DBG2, "[MUXER_OGG] Initializing Ogg Muxer.\n");
    muxer->cont_new_stream = &oggfile_new_stream;
    muxer->cont_write_chunk = &oggfile_write_chunk;
    muxer->cont_write_header = &oggfile_write_header;
    muxer->cont_write_index = &oggfile_write_index;
    return 1;
}
